import customtkinter
import tkinter as tk
from tkinter import filedialog


class PadimMainFrame(customtkinter.CTk):
    def __init__(self):
        super().__init__()
        customtkinter.set_appearance_mode("System")  # Modes: system (default), light, dark
        customtkinter.set_default_color_theme("blue")  # Themes: blue (default), dark-blue, green
        self.n_rows = 5
        self.n_columns = 10

        self.app = customtkinter.CTk()  # create CTk window like you do with the Tk window
        self.app.geometry("1200x800")
        # self.app.wm_attributes('-fullscreen', True)
        # self.app.state('normal')
        self._self_create_layout()

    def browse_button(self):
        """
        allow user to get a filedialog and select
        :return:
        """
        filename = filedialog.askdirectory()
        # self.train_path.set(filename)
        print(filename)
        self.lbl1.configure(text=filename)

    def _self_create_layout(self):
        self.select_train_button = customtkinter.CTkButton(
            master=self.app,
            text="train path",
            command=self.browse_button
        )
        self.select_train_button.grid(row=0, column=0)

        # self.train_path = customtkinter.StringVar()
        # self.train_path.set('undefined')
        self.lbl1 = customtkinter.CTkLabel(master=self.app, text='undefined')
        # self.lbl1 = tk.Label(master=self.app, text=self.train_path.get())
        self.lbl1.grid(row=0, column=1)

    def mainloop(self):
        self.app.mainloop()



import cv2
import numpy as np


class VisionRendering:
    @staticmethod
    def visualize_binary_im_on_rgb(
            bgr_im: np.ndarray,
            bin_im: np.ndarray,
            alpha=0.5,
            gamma=0.0,
    ) -> np.ndarray:
        """
        visualize a binary image on a rgb image for anomaly segmentation.
        Make sure that height and width of the rgb image and the binary image are matching, otherwise an exception
        is thrown
        :param bgr_im: 3 channel image
        :param bin_im: one channel binary image
        :param alpha: weight sharing within the weight
        :param gamma: additional brightness
        :return:
        """
        if bgr_im.shape[:2] != bin_im.shape[:2]:
            raise Exception('shapes not matching of rgb image and binary image: ', bgr_im.shape, bin_im.shape)
        bin_image_red = np.stack([bin_im, bin_im, bin_im], axis=-1)
        bin_image_red[:, :, 1] = 0
        bin_image_red[:, :, 2] = 0
        visualized = cv2.addWeighted(bgr_im, alpha, bin_image_red, 1 - alpha, gamma)
        return visualized



from data.dataset import PadimDataset
from data.transform import DataTransform
from PadimAD import PadimAnomalyDetector
from config.DTO import PadimADConfig


def main():
    config = PadimADConfig(
        model_name='wide_resnet50_2',
        device='cuda',
        batch_size=8
    )
    path_good_data = r'C:\data\mvtec\bottle\train\good'
    path_anomalous_data = r'C:\data\mvtec\bottle\test\contamination'
    path_anomalous_data = r'C:\data\mvtec\bottle\test'
    ad_detector = PadimAnomalyDetector(config=config)
    good_dataset = PadimDataset(data_path=path_good_data, transform=DataTransform.get_train_transform())
    ad_detector.train_anomaly_detection(dataset=good_dataset)
    ad_dataset = PadimDataset(data_path=path_anomalous_data, transform=DataTransform.get_test_transform())
    ad_detector.detect_anomalies_on_dataset(dataset=ad_dataset, path=r'C:\data\padim\anom')
    ad_detector.detect_anomalies_on_dataset(dataset=good_dataset, path=r'C:\data\padim\good')


if __name__ == '__main__':
    main()



import os
import tkinter
import glob
import threading

import customtkinter
import cv2
import numpy as np
from PIL import Image

from gui.gui_components.train_parameters_widget import TrainParametersWidget
from gui.gui_enums.ad_status import ADStatus
from config.DTO import PadimADConfig
from data.dataset import PadimDataset
from data.transform import DataTransform
from PadimAD import PadimAnomalyDetector
from vision.rendering import VisionRendering


class App(customtkinter.CTk):
    app_mode = None
    assets_path = None
    ad_status = None
    # training frame attributes
    train_status_label = None
    whistle_image = None
    train_im_paths = None
    train_images = None
    n_sample_images = 5
    train_header = None
    train_header_stringvar = None
    n_train_columns = 7
    start_train_button = None
    train_progressbar = None
    train_thread = None
    train_config = None
    ad_detector = None
    train_status_stringvar = None
    # validation attributes within the training frame
    val_path_button = None
    val_path_stringvar = None
    val_path_label = None
    val_im_paths = None
    val_images = None
    val_progressbar = None
    start_val_calibration_button = None
    calibrate_image = None
    calibration_thread = None
    # inspection frame attributes
    inspection_frame = None
    inspect_header_stringvar = None
    inspect_header = None
    inspection_path_button = None
    inspection_im_path = None
    inspection_ctk_im = None
    inspection_im_panel = None
    detect_anomalies_button = None
    insp_result_ctk_im = None
    insp_result_im_panel = None
    slider = None
    slider_value = None
    insp_result_bin_image_ctk_im = None
    insp_result_bin_im_panel = None
    insp_result_vis_image_ctk_im = None
    insp_result_vis_im_panel = None

    def __init__(self):
        super().__init__()
        # status variables
        # self.temp_var = ADStatus.not_calibrated
        self.train_parameters_widget = None
        self.ad_status = ADStatus.not_calibrated

        # variables instantiated later on
        self.train_path_stringvar = None
        self.train_path_label = None
        self.train_path_button = None

        # constructor code
        self.title("anomaly_detection_ui.py")
        self.geometry("1400x800")

        # set grid layout 1x2
        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(1, weight=1)

        # load images with light and dark mode image
        self.assets_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "images")
        self.logo_image = customtkinter.CTkImage(
            Image.open(os.path.join(self.assets_path, "anomaly.png")), size=(26, 26))
        self.large_test_image = customtkinter.CTkImage(
            Image.open(os.path.join(self.assets_path, "logo.png")), size=(500, 150))
        self.image_icon_image = customtkinter.CTkImage(
            Image.open(os.path.join(self.assets_path, "logo.png")), size=(20, 20))
        self.file_folder_image = customtkinter.CTkImage(
            Image.open(os.path.join(self.assets_path, "file-and-folder.png")), size=(20, 20))

        self.home_image = customtkinter.CTkImage(
            light_image=Image.open(os.path.join(self.assets_path, "home.png")),
            dark_image=Image.open(os.path.join(self.assets_path, "home.png")),
            size=(20, 20),
        )
        self.training_image = customtkinter.CTkImage(
            light_image=Image.open(os.path.join(self.assets_path, "ai.png")),
            dark_image=Image.open(os.path.join(self.assets_path, "ai.png")),
            size=(20, 20),
        )
        self.inspect = customtkinter.CTkImage(
            light_image=Image.open(os.path.join(self.assets_path, "analysis.png")),
            dark_image=Image.open(os.path.join(self.assets_path, "analysis.png")),
            size=(20, 20),
        )

        # create navigation frame
        self.navigation_frame = customtkinter.CTkFrame(self, corner_radius=0)
        self.navigation_frame.grid(row=0, column=0, sticky="nsew")
        self.navigation_frame.grid_rowconfigure(4, weight=1)

        self.navigation_frame_label = customtkinter.CTkLabel(
            self.navigation_frame,
            text="Anomaly Detector",
            image=self.logo_image,
            compound="left",
            font=customtkinter.CTkFont(size=15, weight="bold"),
        )
        self.navigation_frame_label.grid(row=0, column=0, padx=20, pady=20)

        self.home_button = customtkinter.CTkButton(
            self.navigation_frame, corner_radius=0, height=40, border_spacing=10, text="Home",
            fg_color="transparent", text_color=("gray10", "gray90"), hover_color=("gray70", "gray30"),
            image=self.home_image, anchor="w", command=self.home_button_event)
        self.home_button.grid(row=1, column=0, sticky="ew")

        self.frame_2_button = customtkinter.CTkButton(
            self.navigation_frame, corner_radius=0, height=40, border_spacing=10, text="Train",
            fg_color="transparent", text_color=("gray10", "gray90"), hover_color=("gray70", "gray30"),
            image=self.training_image, anchor="w", command=self.frame_2_button_event)
        self.frame_2_button.grid(row=2, column=0, sticky="ew")

        self.frame_3_button = customtkinter.CTkButton(
            self.navigation_frame, corner_radius=0, height=40, border_spacing=10, text="Inspect",
            fg_color="transparent", text_color=("gray10", "gray90"), hover_color=("gray70", "gray30"),
            image=self.inspect, anchor="w", command=self.frame_3_button_event)
        self.frame_3_button.grid(row=3, column=0, sticky="ew")

        self.appearance_mode_menu = customtkinter.CTkOptionMenu(
            self.navigation_frame, values=["Light", "Dark", "System"],
            command=self.change_appearance_mode_event)

        # create home frame
        self.home_frame = customtkinter.CTkFrame(self, corner_radius=0, fg_color="transparent")
        self.home_frame.grid_columnconfigure(0, weight=1)

        self.home_frame_large_image_label = customtkinter.CTkLabel(
            self.home_frame, text="", image=self.large_test_image)
        self.home_frame_large_image_label.grid(row=0, column=0, padx=20, pady=10)

        self.home_frame_button_1 = customtkinter.CTkButton(self.home_frame, text="", image=self.image_icon_image)
        self.home_frame_button_1.grid(row=1, column=0, padx=20, pady=10)
        self.home_frame_button_2 = customtkinter.CTkButton(
            self.home_frame, text="CTkButton", image=self.image_icon_image, compound="right")
        self.home_frame_button_2.grid(row=2, column=0, padx=20, pady=10)
        self.home_frame_button_3 = customtkinter.CTkButton(
            self.home_frame, text="CTkButton", image=self.image_icon_image, compound="top")
        self.home_frame_button_3.grid(row=3, column=0, padx=20, pady=10)
        self.home_frame_button_4 = customtkinter.CTkButton(
            self.home_frame, text="CTkButton", image=self.image_icon_image, compound="bottom", anchor="w")
        self.home_frame_button_4.grid(row=4, column=0, padx=20, pady=10)

        # create second frame
        self.train_frame = customtkinter.CTkFrame(self, corner_radius=0, fg_color="transparent")
        self.train_frame.grid_columnconfigure(0, weight=1)
        self.create_train_frame()

        # create third frame
        self.inspection_frame = customtkinter.CTkFrame(self, corner_radius=0, fg_color="transparent")
        self.create_inspection_frame()

        # select default frame
        self.select_frame_by_name("home")
        # TODO debug - set back to home, for implementation reasons
        self.select_frame_by_name("frame_2")

    def create_train_frame(self):
        for i in range(self.n_train_columns):
            self.train_frame.columnconfigure(i, weight=1)

        self.train_header_stringvar = tkinter.StringVar(value="Train Data")
        self.train_header = customtkinter.CTkLabel(
            master=self.train_frame,
            textvariable=self.train_header_stringvar,
            width=250,
            height=50,
            bg_color="dodger blue",
            corner_radius=16
        )
        self.train_header.grid(row=0, column=0, padx=20, pady=10)

        self.train_status_stringvar = tkinter.StringVar(value=str(self.ad_status))
        self.train_status_label = customtkinter.CTkLabel(
            master=self.train_frame,
            textvariable=self.train_status_stringvar,
            width=250,
            height=50,
            bg_color="dodger blue",
            corner_radius=16
        )
        self.train_status_label.grid(row=0, column=1, padx=20, pady=10)

        self.train_path_button = customtkinter.CTkButton(
            self.train_frame,
            text="Set Train Path",
            image=self.file_folder_image,
            compound="right",
            command=self.select_train_path,
        )
        self.train_path_button.grid(row=1, column=0, padx=20, pady=10)

        self.train_path_stringvar = tkinter.StringVar(value="Not Set")

        self.train_path_label = customtkinter.CTkLabel(
            master=self.train_frame,
            textvariable=self.train_path_stringvar,
            # width=500,
            # height=25,
            bg_color="dodger blue",
            corner_radius=16
        )
        self.train_path_label.grid(row=1, column=1, padx=20, pady=10)
        # create the button that starts training
        self.whistle_image = customtkinter.CTkImage(
            Image.open(os.path.join(self.assets_path, "start_training.png")), size=(20, 20))

        self.start_train_button = customtkinter.CTkButton(
            self.train_frame,
            text="Train",
            image=self.whistle_image,
            compound="right",
            command=self.start_training,
        )
        self.start_train_button.grid(row=2, column=0, padx=10, pady=5)
        self.train_progressbar = customtkinter.CTkProgressBar(
            master=self.train_frame, width=800, corner_radius=5, height=50)
        self.train_progressbar.set(0)
        self.train_progressbar.grid(row=2, column=1, padx=10, pady=5, columnspan=6)
        # create
        self.train_parameters_widget = TrainParametersWidget(self.train_frame, corner_radius=0, fg_color="transparent")
        self.train_parameters_widget.grid(row=6, column=0)
        self.create_validation_frame()

    def create_validation_frame(self):
        """
        creates the validation frame within the training frame
        :return:
        """
        self.val_path_button = customtkinter.CTkButton(
            self.train_frame,
            text="Set Val Path",
            image=self.file_folder_image,
            compound="right",
            command=self.select_val_path,
        )
        self.val_path_button.grid(row=3, column=0, padx=20, pady=10)
        self.val_path_stringvar = tkinter.StringVar(value="Not Set")

        self.val_path_label = customtkinter.CTkLabel(
            master=self.train_frame,
            textvariable=self.val_path_stringvar,
            # width=500,
            # height=25,
            bg_color="dodger blue",
            corner_radius=16
        )
        self.val_path_label.grid(row=3, column=1, padx=20, pady=10)
        self.calibrate_image = customtkinter.CTkImage(
            Image.open(os.path.join(self.assets_path, "kalibrieren.png")), size=(20, 20))
        self.start_val_calibration_button = customtkinter.CTkButton(
            self.train_frame,
            text="Calibrate",
            image=self.calibrate_image,
            compound="right",
            command=self.calibrate_validation_data,
        )
        self.start_val_calibration_button.grid(row=4, column=0, padx=10, pady=5)
        self.val_progressbar = customtkinter.CTkProgressBar(
            master=self.train_frame, width=800, corner_radius=5, height=50)
        self.val_progressbar.set(0)
        self.val_progressbar.grid(row=4, column=1, padx=10, pady=5, columnspan=6)

    def calibrate_validation_data(self):
        """
        calibrate the anomaly detector on the validation data.
        :return:
        """
        val_dataset = PadimDataset(
            data_path=self.val_path_stringvar.get(),
            transform=DataTransform.get_train_transform(
                im_size=self.train_parameters_widget.im_size,
                crop_size=self.train_parameters_widget.im_size,
            )
        )
        self.calibration_thread = threading.Thread(
            target=self.ad_detector.calibrate_anomalies_on_dataset,
            args=(val_dataset, self.val_progressbar)
        )
        self.calibration_thread.start()
        # self.ad_detector.calibrate_anomalies_on_dataset(dataset=val_dataset, progress_bar=self.val_progressbar)
        # print(self.ad_detector.cal_min_score)
        # print(self.ad_detector.cal_max_score)

    def select_val_path(self):
        filepath = tkinter.filedialog.askdirectory()
        self.val_path_stringvar.set(filepath)
        self.init_val_data()

    def init_val_data(self):
        """
        initialize the validation data.
        :return:
        """
        im_paths = glob.glob(os.path.join(self.val_path_stringvar.get(), '**/*.png'), recursive=True)
        print('[INFO] found {} images for training'.format(len(im_paths)))
        self.val_im_paths = im_paths
        self.val_images = []
        for i in range(self.n_sample_images):
            tmp_im = customtkinter.CTkImage(
                light_image=Image.open(self.val_im_paths[i]),
                dark_image=Image.open(self.val_im_paths[i]),
                size=(100, 100),
            )
            im_panel = customtkinter.CTkLabel(
                self.train_frame,
                image=tmp_im,
                text='',
            )
            im_panel.grid(row=3, column=i + 2, padx=20, pady=10)
            self.val_images.append(im_panel)

    def create_inspection_frame(self):
        """
        create inspection frame
        :return:
        """
        self.inspect_header_stringvar = tkinter.StringVar(value="Inspection")
        self.inspect_header = customtkinter.CTkLabel(
            master=self.inspection_frame,
            textvariable=self.inspect_header_stringvar,
            width=250,
            height=50,
            bg_color="dodger blue",
            corner_radius=16
        )
        self.inspect_header.grid(row=0, column=0, padx=20, pady=10)
        self.inspection_path_button = customtkinter.CTkButton(
            self.inspection_frame,
            text="Load Image",
            image=self.file_folder_image,
            compound="right",
            command=self.select_inspection_image,
        )
        self.inspection_path_button.grid(row=1, column=0, padx=20, pady=10)
        # inspection button
        self.detect_anomalies_button = customtkinter.CTkButton(
            self.inspection_frame,
            text="Detect Anomalies",
            image=self.inspect,
            compound="right",
            command=self.detect_anomalies,
        )
        self.detect_anomalies_button.grid(row=2, column=0, padx=20, pady=10)

        self.slider = customtkinter.CTkSlider(self.inspection_frame, from_=0, to=255, command=self.print_slider)
        self.slider.grid(row=3, column=0, padx=(20, 10), pady=(10, 10))
        self.slider_value = customtkinter.CTkLabel(
            self.inspection_frame,
            image=self.inspection_ctk_im,
            text='{}'.format(self.slider.get()),
        )
        self.slider_value.grid(row=3, column=1, padx=20, pady=10)

    def detect_anomalies(self):
        """
        detect anomalies of an image selected in the ui
        :return:
        """
        self.slider.configure(to=255)
        src_im = Image.open(self.inspection_im_path).convert('RGB')
        anom_score = self.ad_detector.detect_anomaly(
            im=src_im,
            transform=DataTransform.get_test_transform(
                im_size=self.train_parameters_widget.im_size,
                crop_size=self.train_parameters_widget.im_size
            ),
            normalize=True,
        )
        anomaly_im = Image.fromarray(anom_score.astype(np.uint8))

        self.insp_result_ctk_im = customtkinter.CTkImage(
            light_image=anomaly_im,
            dark_image=anomaly_im,
            size=(256, 256),
        )
        self.insp_result_im_panel = customtkinter.CTkLabel(
            self.inspection_frame,
            image=self.insp_result_ctk_im,
            text='',
        )
        self.insp_result_im_panel.grid(row=2, column=1, padx=20, pady=10)

        ret, bin_image_arr = cv2.threshold(anom_score.astype(np.uint8), self.slider.get(), 255, cv2.THRESH_BINARY)
        bin_image = Image.fromarray(bin_image_arr)
        self.insp_result_bin_image_ctk_im = customtkinter.CTkImage(
            light_image=bin_image,
            dark_image=bin_image,
            size=(256, 256),
        )
        self.insp_result_bin_im_panel = customtkinter.CTkLabel(
            self.inspection_frame,
            image=self.insp_result_bin_image_ctk_im,
            text='',
        )
        self.insp_result_bin_im_panel.grid(row=2, column=2, padx=20, pady=10)

        # draw visualized version into the frame
        src_im_arr = cv2.imread(self.inspection_im_path)
        bin_image_arr = cv2.resize(
            bin_image_arr.astype(np.uint8),
            (src_im_arr.shape[1], src_im_arr.shape[0]), cv2.INTER_LINEAR_EXACT)
        visualized = VisionRendering.visualize_binary_im_on_rgb(bgr_im=src_im_arr, bin_im=bin_image_arr)
        self.insp_result_vis_image_ctk_im = customtkinter.CTkImage(
            light_image=Image.fromarray(visualized),
            dark_image=Image.fromarray(visualized),
            size=(256, 256),
        )
        self.insp_result_vis_im_panel = customtkinter.CTkLabel(
            self.inspection_frame,
            image=self.insp_result_vis_image_ctk_im,
            text='',
        )
        self.insp_result_vis_im_panel.grid(row=1, column=2, padx=20, pady=10)

    def print_slider(self, value):
        print(value)
        # self.slider.configure(to=300)
        self.slider_value.configure(text=value)
        # value = self.slider.get()

    def select_frame_by_name(self, name):
        # set button color for selected button
        self.home_button.configure(fg_color=("gray75", "gray25") if name == "home" else "transparent")
        self.frame_2_button.configure(fg_color=("gray75", "gray25") if name == "frame_2" else "transparent")
        self.frame_3_button.configure(fg_color=("gray75", "gray25") if name == "frame_3" else "transparent")

        # show selected frame
        if name == "home":
            self.home_frame.grid(row=0, column=1, sticky="nsew")
        else:
            self.home_frame.grid_forget()
        if name == "frame_2":
            self.train_frame.grid(row=0, column=1, sticky="nsew")
        else:
            self.train_frame.grid_forget()
        if name == "frame_3":
            self.inspection_frame.grid(row=0, column=1, sticky="nsew")
        else:
            self.inspection_frame.grid_forget()

    def home_button_event(self):
        self.select_frame_by_name("home")

    def frame_2_button_event(self):
        self.select_frame_by_name("frame_2")

    def frame_3_button_event(self):
        self.select_frame_by_name("frame_3")

    def change_appearance_mode_event(self, new_appearance_mode):
        self.app_mode = new_appearance_mode
        customtkinter.set_appearance_mode(new_appearance_mode)

    def change_ad_status(self, new_status: ADStatus):
        self.ad_status = new_status
        self.train_status_stringvar.set(str(self.ad_status))

    def select_train_path(self):
        """
        select the training path and display it in the GUI, also extract images
        :return:
        """
        filepath = tkinter.filedialog.askdirectory()
        self.train_path_stringvar.set(filepath)
        self.init_train_data()

    def select_inspection_image(self):
        """
        select an inspection image and load it into the GUI
        :return:
        """
        self.inspection_im_path = r'C:\data\mvtec\bottle\test\broken_large\003.png'
        file = tkinter.filedialog.askopenfile()
        self.inspection_im_path = os.path.abspath(file.name)
        print('[INFO] selected inspection image: {}'.format(self.inspection_im_path))
        self.inspection_ctk_im = customtkinter.CTkImage(
            light_image=Image.open(self.inspection_im_path),
            dark_image=Image.open(self.inspection_im_path),
            size=(256, 256),
        )
        self.inspection_im_panel = customtkinter.CTkLabel(
            self.inspection_frame,
            image=self.inspection_ctk_im,
            text='',
        )
        self.inspection_im_panel.grid(row=1, column=1, padx=20, pady=10)

    def init_train_data(self):
        """
        whenever a training path is selected the data is initialized again
        :return:
        """
        im_paths = glob.glob(os.path.join(self.train_path_stringvar.get(), '**/*.png'), recursive=True)
        print('[INFO] found {} images for training'.format(len(im_paths)))
        self.train_im_paths = im_paths
        self.train_images = []
        for i in range(self.n_sample_images):
            tmp_im = customtkinter.CTkImage(
                light_image=Image.open(self.train_im_paths[i]),
                dark_image=Image.open(self.train_im_paths[i]),
                size=(100, 100),
            )
            im_panel = customtkinter.CTkLabel(
                self.train_frame,
                image=tmp_im,
                text='',
            )
            im_panel.grid(row=1, column=i + 2, padx=20, pady=10)
            self.train_images.append(im_panel)

    def run_training_thread(self):
        """

        :return:
        """
        self.train_config = PadimADConfig(
            model_name='wide_resnet50_2',
            device='cuda',
            batch_size=8
        )
        self.ad_detector = PadimAnomalyDetector(config=self.train_config)
        good_dataset = PadimDataset(
            data_path=self.train_path_stringvar.get(),
            transform=DataTransform.get_train_transform(
                im_size=self.train_parameters_widget.im_size,
                crop_size=self.train_parameters_widget.im_size,
            )
        )
        self.change_ad_status(new_status=ADStatus.calibrating)
        self.ad_detector.train_anomaly_detection(dataset=good_dataset, progress_bar=self.train_progressbar)
        self.change_ad_status(new_status=ADStatus.calibrated)

    def start_training(self):
        self.train_thread = threading.Thread(target=self.run_training_thread)
        self.train_thread.start()


if __name__ == "__main__":
    app = App()
    app.mainloop()



import tkinter
import customtkinter
from customtkinter import *


class TrainParametersWidget(CTkFrame):
    """
    padim widget to control certain parameters of the application
    """
    def __init__(self, parent, corner_radius, fg_color):
        super().__init__(parent, corner_radius=corner_radius, fg_color=fg_color)
        # self.base_frame = customtkinter.CTkFrame(self, corner_radius=0)
        self.im_size_label_stringvar = tkinter.StringVar(value="Image Size")
        self.im_size_label = customtkinter.CTkLabel(
            master=self,
            textvariable=self.im_size_label_stringvar,
            width=80,
            height=30,
            bg_color="dodger blue",
            corner_radius=16
        )
        self.im_size_label.grid(row=0, column=0, padx=20, pady=10)
        self.im_size_entry = CTkEntry(self)
        self.im_size_entry.insert(END, '224')
        self.im_size_entry.grid(row=0, column=1, padx=20, pady=10)

    @property
    def im_size(self) -> int:
        return int(self.im_size_entry.get())

        # self.im_w_stringvar = tkinter.StringVar(value="Image Width")
        # self.im_w_label = customtkinter.CTkLabel(
        #     master=self,
        #     textvariable=self.im_w_stringvar,
        #     width=80,
        #     height=30,
        #     bg_color="dodger blue",
        #     corner_radius=16,
        # )
        # self.im_w_label.grid(row=1, column=0, padx=20, pady=10)
        # self.im_w_entry = CTkEntry(self)
        # self.im_w_entry.insert(END, '224')
        # self.im_w_entry.grid(row=1, column=1, padx=20, pady=10)




from random import sample

import torch
from torchvision.models import wide_resnet50_2, resnet18

from constants import constants_ai
from ai.DTO import FeatureExtraction


class FeatureExtractor:
    def __init__(self, model_name: str, device='cuda'):
        """
        create feature extractor with a model backbone
        :param model_name: name of the model to extract features with
        """
        self.model_name = model_name
        self.model = None
        self.t_d = 0
        self.d = 0
        self.device = device
        self.outputs = []
        self.idx = None
        self.build_feature_extraction_model(model_name=model_name)

    def build_feature_extraction_model(self, model_name: str):
        """
        build
        :param model_name:
        :return:
        """
        if model_name == 'resnet18':
            self.model = resnet18(pretrained=True, progress=True)
            self.t_d = 448
            self.d = 100
        elif model_name == 'wide_resnet50_2':
            self.model = wide_resnet50_2(pretrained=True, progress=True)
            self.t_d = 1792
            self.d = 550
        else:
            raise ValueError('no valid model name provided: {}, select from {}'.format(
                model_name,
                constants_ai.supported_feature_extraction_models
            ))
        self.model.to(self.device)
        self.model.eval()

        # set model's intermediate outputs
        def hook(module, inputs, output):
            self.outputs.append(output)

        self.model.layer1[-1].register_forward_hook(hook)
        self.model.layer2[-1].register_forward_hook(hook)
        self.model.layer3[-1].register_forward_hook(hook)
        self.idx = torch.tensor(sample(range(0, self.t_d), self.d))

    def extract_features(self, in_sample: torch.tensor) -> FeatureExtraction:
        """
        extract features from a given sample.
        :param in_sample: sample to extract features from
        :return:
        """
        _ = self.model(in_sample.to(self.device))

        fe = FeatureExtraction(
            layer_0=self.outputs[0].clone(),
            layer_1=self.outputs[1].clone(),
            layer_2=self.outputs[2].clone(),
        )
        self.outputs = []
        return fe



from PIL import Image
import numpy as np
from torchvision import transforms as T


class DataTransform:
    @staticmethod
    def denormalize_image(x):
        mean = np.array([0.485, 0.456, 0.406])
        std = np.array([0.229, 0.224, 0.225])
        x = (((x.transpose(1, 2, 0) * std) + mean) * 255.).astype(np.uint8)

        return x

    @staticmethod
    def get_train_transform(im_size=256, crop_size=224):
        """
        get training transformation for images
        :param im_size: symmetric size to resize the image to
        :param crop_size:  center crop operation size
        :return:
        """
        transform_x = T.Compose([
            T.Resize(im_size, Image.LANCZOS),
            T.CenterCrop(crop_size),
            T.ToTensor(),
            T.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        ])
        return transform_x

    @staticmethod
    def get_test_transform(im_size=256, crop_size=224):
        """
        get transformation for training
        :param im_size:  symmetric size to resize the image to
        :param crop_size: center crop operation size
        :return:
        """
        transform_x = T.Compose([
            T.Resize(im_size, Image.LANCZOS),
            T.CenterCrop(crop_size),
            T.ToTensor(),
            T.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        ])
        return transform_x



from typing import Optional
from dataclasses import dataclass

import torch
import torch.nn.functional as F


@dataclass
class FeatureExtraction:
    layer_0: torch.Tensor
    layer_1: torch.Tensor
    layer_2: torch.Tensor
    embedded_vectors: Optional[torch.Tensor] = None

    @staticmethod
    def embedding_concat(x, y):
        B, C1, H1, W1 = x.size()
        _, C2, H2, W2 = y.size()
        s = int(H1 / H2)
        x = F.unfold(x, kernel_size=s, dilation=1, stride=s)
        x = x.view(B, C1, -1, H2, W2)
        z = torch.zeros(B, C1 + C2, x.size(2), H2, W2)
        for i in range(x.size(2)):
            z[:, :, i, :, :] = torch.cat((x[:, :, i, :, :], y), 1)
        z = z.view(B, -1, H2 * W2)
        z = F.fold(z, kernel_size=s, output_size=(H1, W1), stride=s)

        return z

    def embed_vectors(self):
        """
        embed the vectors into each to form one entire array per pixel in the image
        :return:
        """
        embedding_vectors = self.layer_0
        embedding_vectors = FeatureExtraction.embedding_concat(x=embedding_vectors, y=self.layer_1)
        self.embedded_vectors = FeatureExtraction.embedding_concat(x=embedding_vectors, y=self.layer_2)

    def detach_cpu(self):
        self.layer_0 = self.layer_0.cpu().detach()
        self.layer_1 = self.layer_1.cpu().detach()
        self.layer_2 = self.layer_2.cpu().detach()

    def move_to_device(self, device: str):
        self.layer_0 = self.layer_0.to(device)
        self.layer_1 = self.layer_1.to(device)
        self.layer_2 = self.layer_2.to(device)



import os
import glob

from PIL import Image
from torch.utils.data import Dataset


class PadimDataset(Dataset):
    """
    Class to manage data loading for the Padim Anomaly Detection algorithm.
    """
    def __init__(self, data_path: str, transform, file_extensions='png'):
        """
        default constructor for the PadimDataset class
        :param data_path: path to the data
        :param transform: transformations to run on the data
        :param file_extensions: file extension to be used
        """
        self.data_path = data_path
        self.file_extension = file_extensions
        self.data = glob.glob(os.path.join(data_path, '**/*.{}'.format(self.file_extension)), recursive=True)
        self.transform = transform

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        im_path = self.data[item]
        x = Image.open(im_path).convert('RGB')
        x = self.transform(x)
        return x



import random
import os

import cv2
import numpy as np
import pickle
from scipy.ndimage import gaussian_filter
from scipy.spatial.distance import mahalanobis
import torch
from torch.utils.data import DataLoader
import torch.nn.functional as F
from tqdm import tqdm

from ai.DTO import FeatureExtraction
from ai.feature_extraction import FeatureExtractor
from config.DTO import PadimADConfig
from data.transform import DataTransform


class PadimAnomalyDetector:

    train_outputs = None
    cal_min_score = -1  # detected minimum score on the calibrated data
    cal_max_score = -1  # detected maximum score on the calibrated data

    """
    central class for managing to create, load, run and use anomaly detection based on PaDim.
    """
    def __init__(self, config: PadimADConfig):
        """
        default constructor with config for a Padim Anomaly Detector
        :param config:
        """
        self.config = config
        self.feat_extractor = FeatureExtractor(
            model_name=config.model_name,
            device=config.device,
        )
        # TODO debug statement to check results for repeatability
        random.seed(1024)
        torch.manual_seed(1024)
        if config.device == 'cuda':
            torch.cuda.manual_seed_all(1024)

    def train_anomaly_detection(self, dataset, progress_bar=None):
        """
        train an anomaly detection on a given dataset, good data without anomalies
        :param dataset: holds the data on which the AD is being tuned for.
        :param progress_bar: if an ui element is involved provide a progress bar to modify
        :return:
        """
        # TODO pickle is broken here!!
        with open('../train_class.pkl', 'rb') as f:
            self.train_outputs = pickle.load(f)
        # return

        train_dataloader = DataLoader(dataset, batch_size=self.config.batch_size, pin_memory=True)
        print('[INFO] extracting features from training dataset: {}'.format(len(dataset)))
        feature_extractions = []

        aux = 0

        for x in tqdm(train_dataloader, total=len(train_dataloader)):
            fe = self.feat_extractor.extract_features(in_sample=x)
            fe.detach_cpu()
            fe.move_to_device('cpu')
            feature_extractions.append(fe)
            if progress_bar is not None:
                step_width = 1 / len(train_dataloader)
                aux += step_width
                progress_bar.set(aux)
            # progress_bar.step()
        fe_summary = FeatureExtraction(
            layer_0=torch.cat([x.layer_0.clone() for x in feature_extractions], dim=0),
            layer_1=torch.cat([x.layer_1.clone() for x in feature_extractions], dim=0),
            layer_2=torch.cat([x.layer_2.clone() for x in feature_extractions], dim=0),
        )
        fe_summary.embed_vectors()

        # randomly select d dimension
        embedding_vectors = torch.index_select(fe_summary.embedded_vectors, 1, self.feat_extractor.idx)
        # calculate multivariate Gaussian distribution
        B, C, H, W = embedding_vectors.size()
        embedding_vectors = embedding_vectors.view(B, C, H * W)
        mean = torch.mean(embedding_vectors, dim=0).numpy()
        cov = torch.zeros(C, C, H * W).numpy()
        I = np.identity(C)
        print('[INFO] creating covariance matrices for each pixel')
        aux = 0
        if progress_bar is not None:
            progress_bar.stop()
            progress_bar.set(0)

        for i in tqdm(range(H * W), total=H * W):
            cov[:, :, i] = np.cov(embedding_vectors[:, :, i].numpy(), rowvar=False) + 0.01 * I

            if progress_bar is not None and i % 100 == 0:
                step_width = 1 / (H * W)
                aux += step_width * 100
                progress_bar.set(aux)

        # save the temporary training outputs on mean and covariance
        self.train_outputs = [mean, cov]
        conv_inv_list = []
        aux = 0
        if progress_bar is not None:
            progress_bar.stop()
            progress_bar.set(0)

        for i in tqdm(range(H * W), total=H * W):
            conv_inv = np.linalg.inv(self.train_outputs[1][:, :, i])
            conv_inv_list.append(conv_inv)

            if progress_bar is not None and i % 100 == 0:
                step_width = 1 / (H * W)
                aux += step_width * 100
                progress_bar.set(aux)

        # conv_inv_list = np.array(conv_inv_list).transpose(1, 0).reshape(C, H * W)
        # save learned distribution
        self.train_outputs = [mean, cov, conv_inv_list]

        with open('../train_class.pkl', 'wb') as f:
            pickle.dump(self.train_outputs, f)
        print('[INFO] finished adjusting padim anomaly detector.')

    def detect_anomaly(self, im, transform, normalize=False):
        """
        detect anomaly on an image.
        :param im: image to detect anomaly on.
        :param transform: transformations to apply to the image being sent
        :param normalize: normalize the output map
        :return:
        """
        with torch.no_grad():
            x_in = transform(im)
            x_in = torch.unsqueeze(x_in, 0)
            fe = self.feat_extractor.extract_features(in_sample=x_in)
            fe.detach_cpu()
            fe.move_to_device('cpu')

        fe.embed_vectors()
        embedding_vectors = torch.index_select(fe.embedded_vectors, 1, self.feat_extractor.idx)

        # calculate distance matrix
        B, C, H, W = embedding_vectors.size()
        embedding_vectors = embedding_vectors.view(B, C, H * W).numpy()
        dist_list = []
        for i in tqdm(range(H * W), total=H * W):
            mean = self.train_outputs[0][:, i]
            # conv_inv = np.linalg.inv(self.train_outputs[1][:, :, i])
            conv_inv = self.train_outputs[2][i]
            dist = [mahalanobis(sample[:, i], mean, conv_inv) for sample in embedding_vectors]
            dist_list.append(dist)

        dist_list = np.array(dist_list).transpose(1, 0).reshape(B, H, W)
        # up-sample
        dist_list = torch.tensor(dist_list)
        score_map = F.interpolate(
            dist_list.unsqueeze(1),
            size=x_in.size(2),
            mode='bilinear',
            align_corners=False
        ).squeeze().numpy()

        # apply gaussian smoothing on the score map
        for i in range(score_map.shape[0]):
            score_map[i] = gaussian_filter(score_map[i], sigma=4)

        # Normalization
        max_score = self.cal_max_score
        min_score = self.cal_min_score
        print('max detected on anomaly: {}, max detected on cal: {}'.format(np.max(score_map), max_score))
        print('min detected on anomaly: {}, min detected on cal: {}'.format(np.min(score_map), min_score))
        max_score = max(self.cal_max_score, np.max(score_map))
        test_score_map = (score_map - min_score) / (max_score - min_score)

        if normalize:
            test_score_map = (test_score_map * 255).astype(np.uint8)
        return test_score_map

    def calibrate_anomalies_on_dataset(self, dataset, progress_bar=None):
        """
        calibrate the anomaly detector on a dataset to extract maximum and minimum score.
        The calibration dataset should not contain any anomalies in it, because that would cause the anomaly detector
        to corrupt.
        :param dataset: validation data to calibrate on
        :param progress_bar: if using the gui, use a progress bar to show the progress to the end user
        :return:
        """
        test_dataloader = DataLoader(dataset, batch_size=self.config.batch_size, pin_memory=True)

        test_images = []
        test_fes = []
        aux = 0
        if progress_bar is not None:
            progress_bar.stop()
            progress_bar.set(0)

        for x in tqdm(test_dataloader, '| feature extraction | test |'):
            test_images.extend(x.cpu().detach().numpy())
            with torch.no_grad():
                fe = self.feat_extractor.extract_features(in_sample=x)
                fe.detach_cpu()
                fe.move_to_device('cpu')
                test_fes.append(fe)

            if progress_bar is not None:
                step_width = 1 / len(test_dataloader)
                aux += step_width
                progress_bar.set(aux)

        fe_summary = FeatureExtraction(
            layer_0=torch.cat([x.layer_0.clone() for x in test_fes], dim=0),
            layer_1=torch.cat([x.layer_1.clone() for x in test_fes], dim=0),
            layer_2=torch.cat([x.layer_2.clone() for x in test_fes], dim=0),
        )
        fe_summary.embed_vectors()
        # randomly select d dimension
        embedding_vectors = torch.index_select(fe_summary.embedded_vectors, 1, self.feat_extractor.idx)

        # calculate distance matrix
        B, C, H, W = embedding_vectors.size()
        embedding_vectors = embedding_vectors.view(B, C, H * W).numpy()
        dist_list = []
        aux = 0
        if progress_bar is not None:
            progress_bar.stop()
            progress_bar.set(0)
        for i in tqdm(range(H * W), total=H * W):
            mean = self.train_outputs[0][:, i]
            # conv_inv = np.linalg.inv(self.train_outputs[1][:, :, i])
            conv_inv = self.train_outputs[2][i]
            dist = [mahalanobis(sample[:, i], mean, conv_inv) for sample in embedding_vectors]
            dist_list.append(dist)

            if progress_bar is not None and i % 100 == 0:
                step_width = 1 / (H * W)
                aux += step_width * 100
                progress_bar.set(aux)

        dist_list = np.array(dist_list).transpose(1, 0).reshape(B, H, W)

        # upsample
        dist_list = torch.tensor(dist_list)
        score_map = F.interpolate(
            dist_list.unsqueeze(1),
            size=x.size(2),
            mode='bilinear',
            align_corners=False
        ).squeeze().numpy()

        # apply gaussian smoothing on the score map
        for i in range(score_map.shape[0]):
            score_map[i] = gaussian_filter(score_map[i], sigma=4)

        # Normalization
        max_score = score_map.max()
        min_score = score_map.min()

        self.cal_max_score = max_score * 2
        self.cal_min_score = min_score

    def detect_anomalies_on_dataset(self, dataset, path: str):
        """
        detect anomalies on an entire dataset
        :param dataset: test-dataset to detect anomalies on.
        :param path: write results into a path
        :return:
        """
        test_dataloader = DataLoader(dataset, batch_size=self.config.batch_size, pin_memory=True)

        test_images = []
        test_fes = []
        for x in tqdm(test_dataloader, '| feature extraction | test |'):
            test_images.extend(x.cpu().detach().numpy())
            with torch.no_grad():
                fe = self.feat_extractor.extract_features(in_sample=x)
                fe.detach_cpu()
                fe.move_to_device('cpu')
                test_fes.append(fe)

        fe_summary = FeatureExtraction(
            layer_0=torch.cat([x.layer_0.clone() for x in test_fes], dim=0),
            layer_1=torch.cat([x.layer_1.clone() for x in test_fes], dim=0),
            layer_2=torch.cat([x.layer_2.clone() for x in test_fes], dim=0),
        )
        fe_summary.embed_vectors()
        # randomly select d dimension
        embedding_vectors = torch.index_select(fe_summary.embedded_vectors, 1, self.feat_extractor.idx)

        # calculate distance matrix
        B, C, H, W = embedding_vectors.size()
        embedding_vectors = embedding_vectors.view(B, C, H * W).numpy()
        dist_list = []
        for i in tqdm(range(H * W), total=H * W):
            mean = self.train_outputs[0][:, i]
            conv_inv = np.linalg.inv(self.train_outputs[1][:, :, i])
            dist = [mahalanobis(sample[:, i], mean, conv_inv) for sample in embedding_vectors]
            dist_list.append(dist)

        dist_list = np.array(dist_list).transpose(1, 0).reshape(B, H, W)

        # upsample
        dist_list = torch.tensor(dist_list)
        score_map = F.interpolate(
            dist_list.unsqueeze(1),
            size=x.size(2),
            mode='bilinear',
            align_corners=False
        ).squeeze().numpy()

        # apply gaussian smoothing on the score map
        for i in range(score_map.shape[0]):
            score_map[i] = gaussian_filter(score_map[i], sigma=4)

        # Normalization
        max_score = self.cal_max_score
        min_score = self.cal_min_score
        scores = (score_map - min_score) / (max_score - min_score)
        score_diff = scores.max() - scores.min()
        threshold = int(0.356 * 255)
        for i in range(len(test_images)):
            origin_im = DataTransform.denormalize_image(test_images[i])
            cv2.imwrite(os.path.join(path, '{:02d}_origin_im.png'.format(i)), origin_im)
            test_score_map = (scores[i] * 255).astype(np.uint8)
            cv2.imwrite(os.path.join(path, '{:02d}_scores.png'.format(i)), test_score_map)
            (T, thresh) = cv2.threshold(test_score_map, threshold, 255, cv2.THRESH_BINARY)
            cv2.imwrite(os.path.join(path, '{:02d}_thresh.png'.format(i)), thresh)

